(define-module (sx-ext e-lib-write))
(use-modules  (ice-9 regex)
              (sx undo) (sx property) (sx basic) 
              (sx selection-basic) (sx view) (sx render)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)

(export command-e-lib-write)


; add the commands to the sx system
(entry-add-command-list '("e-lib-write"))



(define (command-e-lib-write)
  (selection-setup selection-string
                   operation-e-lib-write
                   "Enter filename (with path, no extension):"
                   *OPTION_EMPTY* 'view #f #f)
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; private functions

; writes the header file of the e-lib format 
(define (e-lib-write-header fp)
  (display "TITLE              \"Generated by SX\"\n" fp)
  (display "NMESHPOINTS        "fp)
  (display (length *POINTS*) fp)(newline fp)
  (display "NNODES             "fp)
  (display (length *POINTS*) fp)(newline fp)
  (if (> (length *LINKS*) 0)
    (begin
      (display "NELEMENTS_LINK1    "fp)
      (display (length *LINKS*) fp)(newline fp)
    )
  )
  (if (> (length *TRIANGS*) 0)
    (begin
      (display "NELEMENTS_TRIANG1  "fp)
      (display (length *TRIANGS*) fp)(newline fp)
    )
  )
  (if (> (length *QUADS*) 0)
    (begin
      (display "NELEMENTS_QUAD1    "fp)
      (display (length *QUADS*) fp)(newline fp)
    )
  )
  (newline fp)(newline fp)
)


; writes the meshpoints
(define (e-lib-write-meshpoints fp)
  (display "MESHPOINT_COORDINATES\n" fp)
  (let*
    ( (n 1) (co #f)
      (point-map (map (lambda (x)
                        (display n fp)(display "  " fp)
                        (set! n (1+ n))
                        (set! co (object-property-get 'p (car x) 'coords))
                        (display (car   co) fp)(display "  " fp)
                        (display (cadr  co) fp)(display "  " fp)
                        (display (caddr co) fp)(newline fp)
                        (car x)
                      )
                      *POINTS*
                 )
      )
    )
    (newline fp)(newline fp)
    point-map
  )
)


; writes the elements, link, triangles, quads
(define (e-lib-write-elements point-map fp)
  (set! point-map (reverse point-map))
  ; links
  (if (> (length *LINKS*) 0)
    (let* ( (n 1) (nodes #f) )
      (display "NODES_LINK1\n"fp)
      (for-each (lambda (x)
          (display n fp)(display "  " fp)
          (set! n (1+ n))
          (set! nodes (object-property-get 'l (car x) 'nodes))
          (display (length (member (car nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadr nodes) point-map)) fp)
          (newline fp)
        )
        *LINKS*
      )
      (newline fp)(newline fp)
    )
  )
  ; triangles
  (if (> (length *TRIANGS*) 0)
    (let* ( (n 1) (nodes #f) )
      (display "NODES_TRIANG1\n"fp)
      (for-each (lambda (x)
          (display n fp)(display "  " fp)
          (set! n (1+ n))
          (set! nodes (object-property-get 't (car x) 'nodes))
          (display (length (member (car nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadr nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (caddr nodes) point-map)) fp)
          (newline fp)
        )
        *TRIANGS*
      )
      (newline fp)(newline fp)
    )
  )
  ; quads
  (if (> (length *QUADS*) 0)
    (let* ( (n 1) (nodes #f) )
      (display "NODES_QUAD1\n"fp)
      (for-each (lambda (x)
          (display n fp)(display "  " fp)
          (set! n (1+ n))
          (set! nodes (object-property-get 'q (car x) 'nodes))
          (display (length (member (car nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadr nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (caddr nodes) point-map)) fp)
          (display "  " fp)
          (display (length (member (cadddr nodes) point-map)) fp)
          (newline fp)
        )
        *QUADS*
      )
      (newline fp)(newline fp)
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (operation-e-lib-write)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*Saving in e-lib format is cancelled*")
    )
    ( (string)
      (let*
        ( (filename (string-append *SELECT_ANSWER* ".mdf"))
          (fp (catch #t 
                     (lambda () (open-output-file filename))
                     (lambda args #f)))
        )
        (if fp
          (let* ( (point-map #f) )
            ; write the header
            (e-lib-write-header fp)
            ; write mesh points
            (set! point-map (e-lib-write-meshpoints fp))
            ; write elements
            (e-lib-write-elements point-map fp)
            ; close the file
            (close-output-port fp)
          )
          (gui-entry-add-history
            (string-append "\nERROR: File \"" filename 
                           "\" could not be opened for writing!")
          )
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)




