(define-module (sx-ext gen))
(use-modules  (ice-9 rdelim) (sx selection-basic) (sx selection)
              (sx property) (sx basic) (sx undo)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)

(export command-edge command-triang-side command-quad-side
        command-write-gen)


; add the commands to the sx system
(entry-add-command-list '("edge" "triang-side" "quad-side"
                          "write-gen"))

; --------------------------------------------------------------------

(define (command-edge)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-edge-add
                   "Select edge point(s):" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point no_multiple)
                   #f)
)

(define (command-triang-side)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-triang-side-add
                   "Select triangle edge:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select link
                   '(no_empty type_link no_multiple)
                   #f)
)


(define (command-quad-side)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-quad-side-add
                   "Select quad edge:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select link
                   '(no_empty type_link no_multiple)
                   #f)
)


(define (command-write-gen)
  (selection-setup selection-string
                   operation-write-gen
                   "Enter filename (with path):"
                   *OPTION_EMPTY* 'view #f #f)
)


; --------------------------------------------------------------------


(define (operation-edge-add)
  (if (= (length *SELECT_LIST*) 2)
    (let* ( (id #f) )
      (undo-add-op-mark)
      ; add the edge
      (set! id (link-add-basic #f (list (cadar *SELECT_LIST*) 
                                        (cadadr *SELECT_LIST*)) #f #t))
      (object-property-set 'l id 'edge '(integer 1) #t)

      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Two points are required*")
  )
)



(define (operation-triang-side-add)
  (if (= (length *SELECT_LIST*) 3)
    (let* 
      ( (nodes (closed-polygon? *SELECT_LIST*)) )
      (if nodes
        (let*
          ( (prop (list (cons 'child *SELECT_LIST*))) )
          (undo-add-op-mark)
          (triang-add-basic #f nodes prop #t)
        )
      )
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Three link elements are required*")
  )
)


(define (operation-quad-side-add)
  (if (= (length *SELECT_LIST*) 4)
    (let* 
      ( (nodes (closed-polygon? *SELECT_LIST*)) )
      (if nodes
        (let*
          ( (prop (list (cons 'child *SELECT_LIST*))) )
          (undo-add-op-mark)
          (quad-add-basic #f nodes prop #t)
        )
      )
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    (gui-entry-add-history " *Four link elements are required*")
  )
)



; input is a series of link elements: '((l 71) (l 13) (l 45))
; returns #f in the case of an error or the list of nodes which
; defines the closed polygon (3 4 5 6)
(define (closed-polygon? selection)
  (call-with-current-continuation
    (lambda (return)
      (let* 
        ( (links (map (lambda (x) (cadr x)) selection))
          (node-list (map (lambda (x) 
                            (object-property-get (car x) (cadr x) 'nodes))
                          selection))
          (nodes (car node-list))
          (node-list (cdr node-list))
          (elem #f)
        )
        ; validate that the links make up a closed polygon
        (while (not (null? node-list))
          (let loop ( (x node-list) )
            (if (null? x)
              (begin
                (gui-entry-add-history 
                 "\n*ERROR: Not a closed polygon definition*")
                (return #f)
              )
            )
            (set! elem (car x))
            (if (equal? (car elem) (car nodes))
              (begin
                (set! nodes (append (cdr elem) nodes))
                (set! node-list (delete elem node-list))
              )
              (if (equal? (cadr elem) (car nodes))
                (begin
                  (set! nodes (cons (car elem) nodes))
                  (set! node-list (delete elem node-list))
                )
                (loop (cdr x))
              )
            )
          )
        )
        (let* 
          ( (l-nodes (reverse nodes))
            (first   (car l-nodes))
            (l-nodes (reverse (cdr l-nodes)))
          )
          (if (not (equal? first (car l-nodes)))
            (begin
              (gui-entry-add-history 
               "\n*ERROR: Not a closed polygon definition*")
              (return #f)
            )
          )
          l-nodes
        )
      )
    )
  )
)

; --------------------------------------------------------------------
; writing GEN file
; --------------------------------------------------------------------


; write link objects
(define (write-link total node-map fp)
  (do ((i 0 (1+ i))) ((= i *LINK_UNIQUE*))
    (if (vector-ref *LINKS* i)
      (let* 
        ( (elem (vector-ref *LINKS* i))
          (nodes (assoc-ref elem 'nodes))
          ; the correct order of nodes is stored in reverse
          (n1 (cadr nodes))
          (n2 (car nodes))
          (elemtype (assoc-ref elem 'elemtype))
          (ndiv (assoc-ref elem 'ndivision))
        )
        (if (assoc-ref elem 'edge)
          (write-line (string-append "EDGE " 
                                     (number->string total))
                      fp)
          (write-line (string-append "LINK " 
                                     (number->string total))
                      fp)
        )
        ; only linear division
        (write-line (string-append "DEF LINEAR "
                                   (number->string 
                                    (vector-ref node-map n1))
                                   " "
                                   (number->string 
                                    (vector-ref node-map n2)) )
                    fp)
        (if (not (assoc-ref elem 'edge))
          (begin
            ; check elemtype validity
            (if (and elemtype
                     (eq? (car elemtype) 'integer)
                     (number? (cadr elemtype)))
              (set! elemtype (number->string (cadr elemtype)))
              (set! elemtype "1")
            )
            (write-line (string-append "ELEMTYPE " elemtype) fp)
          )
        )

        ; check division validity
        (if (and ndiv
                 (eq? (car ndiv) 'integer)
                 (number? (cadr ndiv)))
          (set! ndiv (number->string (cadr ndiv)))
          (set! ndiv "1")
        )
        (write-line (string-append "NDIVISION " ndiv) fp)

        (write-line "END\n" fp)
        (set! total (1+ total))
      )
    )
  )
  total
)



(define (write-triang total node-map link-map fp)
  (do ((i 0 (1+ i))) ((= i *TRIANG_UNIQUE*))
    (if (vector-ref *TRIANGS* i)
      (let* 
        ( (elem (vector-ref *TRIANGS* i))
          (nodes (assoc-ref elem 'nodes))
          (def (assoc-ref elem 'child))
          (elemtype (assoc-ref elem 'elemtype))
          (ndiv (assoc-ref elem 'ndivision))
        )
        (write-line (string-append "TRIANG " 
                                   (number->string total))
                    fp)
        ; definition of the triangle object
        (if (null? def)
          ; by points
          (let*
            ( (n1 (car nodes))
              (n2 (cadr nodes))
              (n3 (caddr nodes))
            )
            (write-line "DEF POINT3" fp)
            (write-line (string-append (number->string 
                                         (vector-ref node-map n1))
                                       " "
                                       (number->string 
                                         (vector-ref node-map n2))
                                       " "
                                       (number->string 
                                         (vector-ref node-map n3)) )
                        fp)
          )
          ; by edges
          (let*
            ( (l1 (cadr (car def)))
              (l2 (cadr (cadr def)))
              (l3 (cadr (caddr def)))
            )
            (write-line "DEF EDGE3" fp)
            (write-line (string-append (number->string 
                                         (vector-ref link-map l1))
                                       " "
                                       (number->string 
                                         (vector-ref link-map l2))
                                       " "
                                       (number->string 
                                         (vector-ref link-map l3)) )
                        fp)
          )
        )
        ; check elemtype validity
        (if (and elemtype
                 (eq? (car elemtype) 'integer)
                 (number? (cadr elemtype)))
          (set! elemtype (number->string (cadr elemtype)))
          (set! elemtype "1")
        )
        (write-line (string-append "ELEMTYPE " elemtype) fp)

        ; check division validity
        (if (and ndiv
                 (eq? (car ndiv) 'integer)
                 (number? (cadr ndiv)))
          (set! ndiv (number->string (cadr ndiv)))
          (set! ndiv "1")
        )
        (write-line (string-append "NDIVISION " ndiv) fp)

        (write-line "END\n" fp)
        (set! total (1+ total))
      )
    )
  )
  total
)


(define (write-quad total node-map link-map fp)
  (do ((i 0 (1+ i))) ((= i *QUAD_UNIQUE*))
    (if (vector-ref *QUADS* i)
      (let* 
        ( (elem (vector-ref *QUADS* i))
          (nodes (assoc-ref elem 'nodes))
          (def (assoc-ref elem 'child))
          (elemtype (assoc-ref elem 'elemtype))
          (ndiv (assoc-ref elem 'ndivision))
          (divx "1") (divy "1")
        )
        (write-line (string-append "QUAD " 
                                   (number->string total))
                    fp)
        ; definition of the triangle object
        (if (null? def)
          ; by points
          (let*
            ( (n1 (car nodes))
              (n2 (cadr nodes))
              (n3 (caddr nodes))
              (n4 (cadddr nodes))
            )
            (write-line "DEF POINT4" fp)
            (write-line (string-append (number->string 
                                         (vector-ref node-map n1))
                                       " "
                                       (number->string 
                                         (vector-ref node-map n2))
                                       " "
                                       (number->string 
                                         (vector-ref node-map n3))
                                       " "
                                       (number->string 
                                         (vector-ref node-map n4)) )
                        fp)
          )
          ; by edges
          (let*
            ( (l1 (cadr (car def)))
              (l2 (cadr (cadr def)))
              (l3 (cadr (caddr def)))
              (l4 (cadr (cadddr def)))
            )
            (write-line "DEF EDGE4" fp)
            (write-line (string-append (number->string 
                                         (vector-ref link-map l1))
                                       " "
                                       (number->string 
                                         (vector-ref link-map l2))
                                       " "
                                       (number->string 
                                         (vector-ref link-map l3))
                                       " "
                                       (number->string 
                                         (vector-ref link-map l4)) )
                        fp)
          )
        )
        ; check elemtype validity
        (if (and elemtype
                 (eq? (car elemtype) 'integer)
                 (number? (cadr elemtype)))
          (set! elemtype (number->string (cadr elemtype)))
          (set! elemtype "1")
        )
        (write-line (string-append "ELEMTYPE " elemtype) fp)

        ; divisions are necessary only when it is defined by points
        (if (null? def)
          (begin
            ; check division validity
            (if (and ndiv
                     (eq? (car ndiv) 'integer-list)
                     (number? (cadr ndiv))
                     (number? (caddr ndiv)) )
              (begin
                (set! divx (number->string (cadr ndiv)))
                (set! divy (number->string (caddr ndiv)))
              )
            )
            (write-line (string-append "NDIVISION " divx " " divy) fp)
          )
        )

        (write-line "END\n" fp)
        (set! total (1+ total))
      )
    )
  )
  total
)


(define (operation-write-gen)
  (case *SELECT_STATE*
    ( (cancel)
      ; default option
      (gui-entry-add-history "\n*Saving cancelled*")
    )
    ( (string)
      (let*
        ( (fp (catch #t (lambda () (open-output-file *SELECT_ANSWER*))
                                   (lambda args #f))) )
        (if fp
          (let*
            ( (n 0)
              (elem #f) (total 0) (coords #f) (total-elem 0)
              (node-map (make-vector *POINT_UNIQUE* -1))
              (link-map (make-vector *LINK_UNIQUE* -1))
            )
            ; count the number of objects
            (for-each
              (lambda (x y)
                (do ((i 0 (1+ i))) ((= i x))
                  (if (vector-ref y i)
                    (set! n (1+ n))
                  )
                )
              )
              (list *LINK_UNIQUE* *TRIANG_UNIQUE* *QUAD_UNIQUE*)
              (list *LINKS* *TRIANGS* *QUADS*)
            )
           ; write header section
            (write-line "# Generated by SX" fp)
            (write-line (string-append "NPOINTS  " 
                                       (number->string *POINT_UNIQUE*) )
                        fp)
            (write-line (string-append "NOBJECTS " (number->string n)) fp)

            ; map the points
            (do ((i 0 (1+ i))) ((= i *POINT_UNIQUE*))
              (if (vector-ref *POINTS* i)
                (begin
                  (vector-set! node-map i total)
                  (set! total (1+ total))
                )
              )
            )
            ; write points, coordinates
            (write-line "\nCOORDINATES" fp)
            (do ((i 0 (1+ i))) ((= i *POINT_UNIQUE*))
              (set! elem (vector-ref *POINTS* i))
              (if elem
                (let* ((coords (assoc-ref elem 'coords)))
                  (write-line 
                    (string-append (number->string (vector-ref node-map i))
                                   "  "
                                   (number->string (car coords))
                                   "  "
                                   (number->string (cadr coords))
                                   "  "
                                   (number->string (caddr coords))) 
                    fp)
                )
              )
            )
            (write-line "END\n" fp)


            ; map the links
            (set! total 0)
            (do ((i 0 (1+ i))) ((= i *LINK_UNIQUE*))
              (if (vector-ref *LINKS* i)
                (begin
                  (vector-set! link-map i total)
                  (set! total (1+ total))
                )
              )
            )
            ; write link objects
            (set! total-elem (write-link total-elem node-map fp))

            ; write triangle objects
            (set! total-elem (write-triang total-elem node-map link-map fp))
            ; write quad objects
            (set! total-elem (write-quad total-elem node-map link-map fp))
            
            (close-output-port fp)
          )
          (gui-entry-add-history 
            (string-append "\nERROR: Could not open \""
                           *SELECT_ANSWER*
                           "\" file for writing !"))
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)
